#import "gdb.h"
#import "GdbController.h"
#import <demangle.h>
#import <remote/NXProxy.h>
#import <objc/error.h>
#import <stdio.h>
#import <stdlib.h>
#import <sys/signal.h>
#define DIE *(char *)0 = 1;

extern void execute_command ();
extern RegionManager *regionManager;
extern cthread_t gdb_thread;    
extern int do_display_lines;
extern const char *NXGetDefaultValue(const char *owner, const char *name);

GdbController *controller, *remote_controller;
static char gdbHost[128];

@implementation GdbController

static void badRemoteMethod(NXHandler *handler, const char *function)
{
    NXRemoteException exception = (NXRemoteException)handler->code;
    const char *exceptionStrings[] = {
	"couldn't send message",
	"couldn't receive message",
	"couldn't decode message",
	"unknown method",
	"object inaccessible",
	"unknown exception",
	"object not available",
	"remote internal error",
	"multithreaded recursion deadlock",
	"destination invalid",
	"originator invalid",
	"send timed out",
	"receive timed out",
    };
    const int numExceptionStrings = sizeof(exceptionStrings) / sizeof(*exceptionStrings);
    const char *describeErrors = NXGetDefaultValue("GdbClient", "DescribeErrors");

    if (do_display_lines) {
	do_display_lines = NO;
	remote_controller = controller = nil;
	if (describeErrors && (strcmp(describeErrors, "Yes") == 0)) {
	    printf("Distributed Object exception %d (%s) in %s.\n",
		    exception,
		    (NX_couldntSendException <= exception
			&& exception <= (NX_couldntSendException + numExceptionStrings))
		    ? exceptionStrings[exception - NX_couldntSendException] 
		    : "Unknown exception",
		    function);
	} else
	    printf("Problem communicating with Editor.\n");
    }
}

-init
{
//    [NXConnection debug: "GdbControl"];
    mainThread = thread_self();
    [Value setDataGetter: self];
    [Value setErrorDelegate: self];
    [Type setTypeGetter: self];
    [Block setBlockGetter: self];
    types = NXCreateMapTable(NXPtrValueMapPrototype, 0);
    values = NXCreateMapTable(NXPtrValueMapPrototype, 0);
    blocks = NXCreateMapTable(NXPtrValueMapPrototype, 0);
    demangledNames = NXCreateMapTable(NXStrValueMapPrototype, 0);
    [self allocFrames: 16];
    gethostname(gdbHost, sizeof(gdbHost));
    controller = self;
    
    clientLock = mutex_alloc();
    return self;
}

-(void)allocFrames: (int)nFrames
{
    Class frameClass = [Frame class];
    int i;
    
    if (numFramesAlloced)
	frames = realloc(frames, nFrames * sizeof(Frame));
    else
	frames = malloc(nFrames * sizeof(Frame));
    for (i = numFramesAlloced; i < nFrames; i++) 
        *(Class *)(frames + i) = frameClass;
    numFramesAlloced = nFrames;
}

-(Frame *)frameAt: (int)frameIndex
     functionName: (out char **)pFunctionName
             from: (out unsigned int *)pFrom
{
    Frame *frame = frames + frameIndex;
    
    *pFunctionName = [frame functionName];
    *pFrom = [frame frame];
    return frame;
}

-(Frame *)frameAt: (int)frameIndex
             from: (out unsigned int *)pFrom
{
    Frame *frame = frames + frameIndex;

    *pFrom = [frame frame];
    return frame;
}

-(char *)functionNameAt: (int)frameIndex
{
    Frame *frame = frames + frameIndex;

    return [frame functionName];
}

-(char *)functionNameAt: (int)frameIndex hasSymbols: (int *)hasSymbols
{
    Frame *frame = frames + frameIndex;

    *hasSymbols = [frame hasSymbols];
    return [frame functionName];
}

-(void)disconnect: (BOOL)tellClient
{
    if (connection) {
	NXConnection *aConnection;
	int count;
	NXConnection *remoteConnection;
NX_DURING
	if (tellClient)
	    [client disconnect];	
NX_HANDLER
/* It doesn't do us any good to give an error, we were disconnecting anyway. */
	;
NX_ENDHANDLER
    }
    client = nil;
    while (connection = [connection free]);
    remote_controller = nil;
    while (interruptConnection = [interruptConnection free]);
}

-senderIsInvalid: sender
{
    turn_off_viewing();
    [self disconnect: NO];
    return self;
}

-(void)connectToSpeaker: (Speaker *)speaker
{
    if (speaker) {
	NXPort *listenPort;
	BOOL gotToEnd = NO;
	
	objc_setMultithreaded(YES);
	connection = [NXConnection registerRoot: self];
NX_DURING	
	[connection runInNewThread];
	
	listenPort = [connection inPort];
	remote_controller
	= (GdbController *)[NXConnection connectToPort: listenPort];
	[(NXProxy *)remote_controller
	    setProtocolForProxy: @protocol(GdbInternal)];
	interruptConnection = [NXConnection registerRoot: self];
	[interruptConnection runInNewThread];
	gotToEnd = YES;
NX_HANDLER	
	badRemoteMethod(&NXLocalHandler, "connectToSpeaker:");
NX_ENDHANDLER
	if (gotToEnd) {
	    clientCond = condition_alloc();
	    [speaker selectorRPC: "connectTo:" paramTypes: "s",
		[listenPort machPort]];
	    mutex_lock(clientLock);
	    while (!client)
		condition_wait(clientCond, clientLock);
	    mutex_unlock(clientLock);
	    condition_free(clientCond);
	    clientCond = (condition_t)NULL;
	}
    }
}

static void
noop (foo)
     int foo;
{
}

static void
execute_command_no_tty(char *command)
{
    struct cleanup *old_chain = make_cleanup (noop, 0);
    
    execute_command(command, NO);
    bpstat_do_actions (&stop_bpstat);
    do_cleanups (old_chain);
}

-(void)executeCommand: (char *)command
{
    char tmpCommand[strlen(command) + 1];
    int wasError;
    
    strcpy(tmpCommand, command);
    lock_execute_command(YES);
    thread_suspend(mainThread);
    rl_deprepare_terminal();
    catch_errors_was_error(execute_command_no_tty, tmpCommand, &wasError);
    if (wasError) {
	printf("Error occured processing command from Edit.\n");
	[self programStopped];
    }
    fflush(stdout);
    fflush(stderr);
    rl_prepare_terminal();
    thread_resume(mainThread);
    lock_execute_command(NO);
}

-(void)breakInFile: (char *)filename atLine: (int)line
{
    char *command = alloca(strlen(filename) + 20);
    sprintf(command, "break %s:%d", filename, line);
    [self executeCommand: command];
}

-(void)untilInFile: (char *)filename atLine: (int)line
{
    char *command = alloca(strlen(filename) + 20);
    
    if (inferior_task) {
	sprintf(command, "until %s:%d", filename, line);
	[self executeCommand: command];
    } else {
	sprintf(command, "tbreak %s:%d", filename, line);
	[self executeCommand: command];
	[self executeCommand: "run"];
    }
}

-(void)step
{
    [self executeCommand: "step"];
}

-(void)next
{
    [self executeCommand: "next"];
}

-(void)continue
{
    [self executeCommand: "continue"];
}

-(void)up
{
    [self executeCommand: "up"];
}

-(void)frame
{
    [self executeCommand: "frame"];
}

static void selectFrame(int level)
{
    int i;
    FRAME frame;
    
    for (i = 0, frame = get_prev_frame(NULL);
         i < level;
	 i++, frame = get_prev_frame(frame));
    
    select_frame (frame, level);
    print_sel_frame(0);
}

-(void)selectFrame: (int)frame
{
    int wasError;
    
    lock_execute_command(YES);
    thread_suspend(mainThread);
    rl_deprepare_terminal();
    catch_errors_was_error(selectFrame, numFrames - (frame + 1), 
                           &wasError);
    if (wasError) {
	printf("Error occured processing command from Edit.\n");
	[self programStopped];
    }
    fflush(stdout);
    fflush(stderr);
    rl_prepare_terminal();
    thread_resume(mainThread);
    lock_execute_command(NO);
}

-(void)down
{
    [self executeCommand: "down"];
}

-(void)finish
{
    [self executeCommand: "finish"];
}

-(void)run
{
    extern int caution;
    int oldCaution = caution;
    
    caution = 0;
    [self executeCommand: "run"];
    caution = oldCaution;
}

-(void)quit
{
    extern int caution;
    int oldCaution = caution;

    caution = 0;
    [self executeCommand: "quit"];
    caution = oldCaution;
}

-(void)interrupt
{
    inferiorSigInt();
}

-(void)printExpression: (char *)selection
{
    char *command = alloca(strlen(selection) + 10), *sel;
    
    printf("Printing ");
    printf(selection);
    printf(".\n");
    for (sel = selection; *sel; sel++)
	if (*sel == '\n')
	    *sel = '\\';
    sprintf(command, "print %s", selection);
    [self executeCommand: command];
}

-(void)printExpressionIndirect: (char *)selection
{
    char *command = alloca(strlen(selection) + 10), *sel;
    
    printf("Printing *(");
    printf(selection);
    printf(").\n");
    for (sel = selection; *sel; sel++)
	if (*sel == '\n')
	    *sel = '\\';
    sprintf(command, "print *(%s)", selection);
    [self executeCommand: command];
}

-(unsigned)getScalerValueFrom: (unsigned)from at: (Location *)location withSize: (int)size
{
    unsigned val = 0;
    value theVal;

    switch (location->locType) {
	case LTYPE_ABSOLUTE: case LTYPE_CONST_ADDRESS: case LTYPE_RELATIVE:
	case LTYPE_RELATIVE_BIT: case LTYPE_RELATIVE_ADDRESS: {
	    void *hisAddr;
	    
	    switch (location->locType) {
		case LTYPE_ABSOLUTE: case LTYPE_CONST_ADDRESS:
		    hisAddr = location->locPlace.address;
		    break;
		case LTYPE_RELATIVE:
		    hisAddr = (void *)(from + location->locPlace.offset);
		    break;
		case LTYPE_RELATIVE_BIT:
		    hisAddr
		    = (void *)(from
		               + ((location->locPlace.locBits.bitPos / 32)
			          * 4));
		    break;
		case LTYPE_RELATIVE_ADDRESS: {
		    void **myAddrAddr
		         = WARP((void **)(from + location->locPlace.offset));
			 
		    if (myAddrAddr)
			hisAddr = *myAddrAddr;
		    else
			hisAddr = NULL;
		} break;
		default:
		    hisAddr = NULL;
	    }
	    if (hisAddr) {
		int *iAddr;
		short *sAddr;
		char *cAddr;
		
		if (size == sizeof(int)) {
		    iAddr = WARP((int *)hisAddr);
		    if (iAddr)
			val = *iAddr;
		} else if (size == sizeof(short)) {
		    sAddr = WARP((short *)hisAddr);
		    if (sAddr)
			val = *sAddr;
		} else {
		    cAddr = WARP((char *)hisAddr);
		    if (cAddr)
			val = *cAddr;
		}
	    }
	}    break;
	case LTYPE_CONST:
	    val = location->locPlace.constant;
	    break;
	case LTYPE_REGISTER: {
	    FRAME fr, foundFrame = NULL;
	    
	    for (fr = get_current_frame();
		 fr && !foundFrame;
		 fr = get_prev_frame(fr)) {
		if (fr->frame == (unsigned)from)
		    foundFrame = fr;
	    }
	    if (foundFrame) {
		theVal = value_from_register(builtin_type_unsigned_long,
					     location->locPlace.regNum,
					     foundFrame);
		if (theVal) 
		    val = value_as_long(theVal);
	    }
	 }   break;
	default:;
    }
    return val;
}

-(double)getDoubleValueFrom: (unsigned)from at: (Location *)location withSize: (int)size
{
    double val = 0.0;
    value theVal;

    switch (location->locType) {
	case LTYPE_ABSOLUTE: case LTYPE_CONST_ADDRESS: case LTYPE_RELATIVE:
	case LTYPE_RELATIVE_ADDRESS: {
	    void *hisAddr;
	    
	    switch (location->locType) {
		case LTYPE_ABSOLUTE: case LTYPE_CONST_ADDRESS:
		    hisAddr = location->locPlace.address;
		    break;
		case LTYPE_RELATIVE:
		    hisAddr = (void *)(from + location->locPlace.offset);
		    break;
		case LTYPE_RELATIVE_ADDRESS: {
		    void **myAddrAddr
		         = WARP((void **)(from + location->locPlace.offset));
			 
		    if (myAddrAddr)
			hisAddr = *myAddrAddr;
		    else
			hisAddr = NULL;
		} break;
		default:
		    hisAddr = NULL;
	    }
	    if (hisAddr) {
		float *fAddr;
		double *dAddr;
		
		if (size == sizeof(float)) {
		    fAddr = WARP((float *)hisAddr);
		    if (fAddr)
			val = (double)*fAddr;
		} else if (size == sizeof(double)) {
		    dAddr = WARP((double *)hisAddr);
		    if (dAddr)
			val = *dAddr;
		}
	    }
	}    break;
	case LTYPE_CONST:
	    val = (double)*(float *)&location->locPlace.constant;
	    break;
	case LTYPE_REGISTER: {
	    FRAME fr, foundFrame = NULL;
	    
	    for (fr = get_current_frame();
		 fr && !foundFrame;
		 fr = get_prev_frame(fr)) {
		if (fr->frame == (unsigned)from)
		    foundFrame = fr;
	    }
	    if (foundFrame) {
		if (size == sizeof(float))
		    theVal = value_from_register(builtin_type_float,
					         location->locPlace.regNum,
					         foundFrame);
		else
		    theVal = value_from_register(builtin_type_double,
					         location->locPlace.regNum,
					         foundFrame);
		if (theVal) 
		   val = value_as_double(theVal);
	    }
	 }   break;
	default:;
    }
    return val;
}

#if 0
-(NXData *)getDataFrom: (unsigned)from at: (Location *)location
                   for: (int)length
{
    int *hisAddr = (int *)[self getAddressFrom: from at: location],
        *myAddr = [regionManager pointerFor: hisAddr withSize: length];
    NXData *data;
	
    if (myAddr)
	data = [[NXData alloc] initWithData: myAddr size: length dealloc: NO];
    else
	data = nil;
    return data;
}
#endif 0

-(int)isDataFrom: (unsigned)from for: (int)length
{
    return [regionManager pointerFor: (void *)from withSize: length]
           ? YES : NO;
}

-(unsigned)getAddressFrom: (unsigned)from at: (Location *)location
{
    unsigned addr, *addrAddr;
    switch (location->locType) {
	case LTYPE_ABSOLUTE: case LTYPE_CONST_ADDRESS:
	    addr = (unsigned)location->locPlace.address;
	    break;
	case LTYPE_RELATIVE:
	    addr = from + location->locPlace.offset;
	    break;
	case LTYPE_RELATIVE_BIT:
	    addr = from + ((location->locPlace.locBits.bitPos / 32) * 4);
	    break;
	case LTYPE_RELATIVE_ADDRESS:
	    addrAddr = WARP((unsigned *)(from + location->locPlace.offset));
	    if (addrAddr)
		addr = *addrAddr;
	    else
		addr = 0;
	    break;
	default:
	    addr = 0;
	    break;
    }
    return addr;
}

-(char *)getStringValueFrom: (unsigned)from at: (Location *)location
{
    char *addr = (char *)[self getScalerValueFrom: from at: location withSize: sizeof(addr)];
    return WARPSTRING(addr);
}

-(char *)getStringValueForArrayFrom: (unsigned)from at: (Location *)location
{
    char *addr = (char *)[self getAddressFrom: from at: location];
    return WARPSTRING(addr);
}

-(char *)getFunctionNameFrom: (unsigned)from at: (Location *)location
{
    char *addr = (char *)[self getScalerValueFrom: from at: location withSize: sizeof(addr)];
    struct minimal_symbol *minSym = lookup_minimal_symbol_by_pc((CORE_ADDR)addr);
    
    return minSym ? minSym->name : NULL;
}

-valueError: (ValueError)valueError
{
    error("Error getting type info, %d.\n"
          "%s.\n",
	  valueError, valueErrorStrings[valueError]);
}

char *nameForType(struct type *type)
{
    if (!type)
	return "<unknown type>";
    
    if (TYPE_NAME(type))
	return TYPE_NAME(type);
    
    switch (TYPE_CODE(type)) {
	case TYPE_CODE_ARRAY:
	case TYPE_CODE_PTR:
	case TYPE_CODE_MEMBER:
	case TYPE_CODE_REF:
	case TYPE_CODE_FUNC:
	case TYPE_CODE_METHOD:
	    return nameForType(TYPE_TARGET_TYPE(type));
	    break;
		
	default:
	    return "<unknown type>";
	    break;
    }
}

-(Type *)makeType: (void *)data
{
    id theType;
    struct type *type = data;
    
    if (theType = NXMapGet(types, type))
	return theType;
    
    switch (TYPE_CODE(type)) {
	case TYPE_CODE_PTR: {
	    struct type *targetType = TYPE_TARGET_TYPE(type), 
	                *targetTargetType;

	    theType = [Pointer new];
	    [theType setSize: TYPE_LENGTH(type)];
	    [theType setLazyType: targetType];
	    if (isObjectType(type))
		[theType isObject: YES];
	    else if (((TYPE_LENGTH(targetType) == 1)
	              && (TYPE_CODE(targetType) == TYPE_CODE_INT))
		     || ((TYPE_CODE(targetType) == TYPE_CODE_ARRAY)
		         && (targetTargetType = TYPE_TARGET_TYPE(targetType))
			 && (TYPE_LENGTH(targetTargetType) == 1)
	                 && (TYPE_CODE(targetTargetType) == TYPE_CODE_INT)))
		[theType isString: YES];
	    else if (TYPE_CODE(targetType) == TYPE_CODE_FUNC)
		[theType isFunc: YES];
	    [theType setIndSize: TYPE_LENGTH(targetType)];
	    if (TYPE_NAME(type))
		[theType setTypeName: TYPE_NAME(type)];
	}   break;
	case TYPE_CODE_ARRAY: {
	    int length = TYPE_LENGTH(type);
	    struct type *targetType = TYPE_TARGET_TYPE(type);
	    
	    theType = [Array new];
	    [theType setSize: length];
	    [theType setNumber: length / TYPE_LENGTH(targetType)];
	    [theType setLazyType: targetType];
	    if ((TYPE_LENGTH(targetType) == 1)
	              && (TYPE_CODE(targetType) == TYPE_CODE_INT))
		[theType isString: YES];
	}   break;
	case TYPE_CODE_ENUM: {
	    int count, i;
	    List *theVals;
	    EnumVal *enumVal;
	    struct field *field;
	    
	    theType = [Enumeration new];
	    [theType setSize: TYPE_LENGTH(type)];
	    count = TYPE_NFIELDS(type);
	    theVals = [List newCount: count];
	    [theType setVals: theVals];
	    for (i = 0, field = TYPE_FIELDS(type);
	         i < count;
		 i++, field++) {
		enumVal = [EnumVal new];
		[enumVal setName: field->name];
		[enumVal setVal: field->bitpos];
		[theVals addObject: enumVal];
	    }
	    [theType setTypeName: nameForType(type)];
	}    break;
	case TYPE_CODE_STRUCT: case TYPE_CODE_UNION: {
	    int count, i;
	    List *theFields;
	    Value *theField;
	    struct field *field;
	    
	    theType = (TYPE_CODE(type) == TYPE_CODE_STRUCT)
	              ? [Structure new]
		      : [Union new];
	    [theType setSize: TYPE_LENGTH(type)];
	    count = TYPE_NFIELDS(type);
	    theFields = [List newCount: count];
	    [theType setFields: theFields];
	    for (i = 0, field = TYPE_FIELDS(type);
	         i < count;
		 i++, field++) {
		theField = [Value new];
		[theField setValueName: field->name];
		if (field->bitsize) {
		    [theField setLocType: LTYPE_RELATIVE_BIT];
		    [theField setLocPlace:
			(field->bitpos << 5) | field->bitsize];
		} else {
		    [theField setLocType: LTYPE_RELATIVE];
		    [theField setLocPlace: field->bitpos / 8];
		}
		[theField setType: [self makeType: field->type]];
		[theFields addObject: theField];
	    }
	    [theType setTypeName: nameForType(type)];
	}    break;
	case TYPE_CODE_FUNC:
	    theType = [Function new];
	    [theType setLazyType: TYPE_TARGET_TYPE(type)];
	    if (TYPE_NAME(type))
		[theType setTypeName: TYPE_NAME(type)];
	    break;
	case TYPE_CODE_INT:
	    theType = [Integer new];
	    [theType setSize: TYPE_LENGTH(type)];
	    [theType setIsSigned: !TYPE_UNSIGNED(type)];
	    break;
	case TYPE_CODE_FLT:
	    theType = [FloatingPoint new];
	    [theType setSize: TYPE_LENGTH(type)];
	    break;
	case TYPE_CODE_VOID:
	    theType = [Void new];
	    [theType setSize: 1];
	    break;
	default:
	    theType = nil;
    }
    if (theType) 
	NXMapInsert(types, type, theType);
    return (Type *)theType;
}

static BOOL isNameChar(char c)
{
    return ((('0' <= c) && (c <= '9'))
            || (('a' <= c) && (c <= 'z'))
	    || (('A' <= c) && (c <= 'Z'))
	    || (c == '_'));
}

static void getName(char *name, char **pName)
{
    char *p;
    int len;
    
    for (p = *pName; isNameChar(*p); p++);
    len = p - *pName;
    memcpy(name, *pName, len);
    name[len] = '\0';
    (*pName) += len;
}

static void getNumber(char *name, char **pName)
{
    char *p;
    int len;
    
    for (p = *pName; isdigit(*p); p++);
    len = p - *pName;
    memcpy(name, *pName, len);
    name[len] = '\0';
    (*pName) += len;
}

-(Type *)getTypeNamed: (char *)typeName from: (unsigned)from
{
    char *pName = typeName, name[1000];
    struct block *block = block_for_pc(from);
    struct type *type;
    struct symbol *sym;
    int number;
    
    while (*pName && !isNameChar(*pName))
	pName++;
    if (strncmp(pName, "struct ", 7) == 0) {
	pName += 7;
	getName(name, &pName);
	if (sym = lookup_symbol(name, block, STRUCT_NAMESPACE, NULL, NULL)) {
	    type = SYMBOL_TYPE(sym);
	    if (TYPE_CODE(type) != TYPE_CODE_STRUCT)
		type = NULL;
	} else
	    type = NULL;
    } else if (strncmp(pName, "union ", 6) == 0) {
	pName += 6;
	getName(name, &pName);
	if (sym = lookup_symbol(name, block, STRUCT_NAMESPACE, NULL, NULL)) {
	    type = SYMBOL_TYPE(sym);
	    if (TYPE_CODE(type) != TYPE_CODE_UNION)
		type = NULL;
	} else
	    type = NULL;
    } else if (strncmp(pName, "enum ", 5) == 0) {
	pName += 5;
	getName(name, &pName);
	if (sym = lookup_symbol(name, block, STRUCT_NAMESPACE, NULL, NULL)) {
	    type = SYMBOL_TYPE(sym);
	    if (TYPE_CODE(type) != TYPE_CODE_ENUM)
		type = NULL;
	} else
	    type = NULL;
    } else if (strncmp(pName, "unsigned", 8) == 0) {
	pName += 8;
	while (*pName && !isNameChar(*pName))
	    pName++;
	if (strncmp(pName, "int", 3) == 0) {
	    pName += 3;
	    type = builtin_type_unsigned_int;
	} else if (strncmp(pName, "long", 4) == 0) {
	    pName += 4;
	    type = builtin_type_unsigned_long;
	} else if (strncmp(pName, "short", 5) == 0) {
	    pName += 5;
	    type = builtin_type_unsigned_short;
	} else if (strncmp(pName, "char", 4) == 0) {
	    pName += 4;
	    type = builtin_type_unsigned_char;
	} else
	    type = builtin_type_unsigned_int;
    } else {
	getName(name, &pName);
	sym = lookup_symbol(name, block, VAR_NAMESPACE, NULL, NULL);
	if (sym && (SYMBOL_CLASS(sym) == LOC_TYPEDEF))
	    type = SYMBOL_TYPE(sym);
	else {
	    type = lookup_primitive_typename(name);
	    if (!type)
		type = (struct type *)typeForClassName(name);
	}
    }
    while (type && *pName) {
	while (*pName && (*pName != '*') && (*pName != '['))
	    pName++;
	if (*pName == '*')
	    type = lookup_pointer_type(type);
	else if (*pName == '[') {
	    pName++;
	    getNumber(name, &pName);
	    if (*pName == ']') {
		number = atoi(name);
		type = create_array_type(type, number);
	    } else
		type = NULL;
	}
	if (*pName)
	    pName++;
    }
    if (type)
	return [self makeType: type];
}

-(Value *)makeValue: (struct symbol *)sym
{
    Value *arg;
    
    if (arg = NXMapGet(values, sym))
	return arg;
	
    arg = [Value new];
    [arg setValueName: SYMBOL_NAME(sym)];
    [arg setLocPlace: SYMBOL_VALUE(sym)];
    switch (SYMBOL_CLASS(sym)) {
	case LOC_CONST:
	    [arg setLocType: LTYPE_CONST];
	    break;
	case LOC_REGPARM: case LOC_REGISTER:
	    [arg setLocType: LTYPE_REGISTER];
	    break;
	case LOC_ARG: case LOC_LOCAL:
	    [arg setLocType: LTYPE_RELATIVE];
	    break;
	case LOC_REF_ARG:
	    [arg setLocType: LTYPE_RELATIVE_ADDRESS];
	    break;
	case LOC_STATIC:
	    [arg setLocType: LTYPE_ABSOLUTE];
	    break;
	case LOC_CONST_BYTES:
	    [arg setLocType: LTYPE_CONST_ADDRESS];
	    break;
	default:;
    }
    [arg setType: [self makeType: SYMBOL_TYPE(sym)]];
    NXMapInsert(values, sym, arg);
    return arg;
}

-(Block *)makeBlockForBlock: (struct block *)block
{
    Block *theBlock, *theSuperBlock;
    struct block *super_block;
    
    if (!block->function && (super_block = BLOCK_SUPERBLOCK(block))) {
	theSuperBlock = NXMapGet(blocks, super_block);
	if (!theSuperBlock)
	    theSuperBlock = [self makeBlockForBlock: super_block];
    } else
	theSuperBlock = nil;
    theBlock = [Block new];   
    [theBlock setStart: BLOCK_START(block)];
    [theBlock setEnd:   BLOCK_END(block)];
    [theBlock setSuperBlock: theSuperBlock];
    [theBlock setData: block];
    NXMapInsert(blocks, block, theBlock);
    return theBlock;
}

-(Block *)makeBlock: (unsigned)pc
{
    Block *theBlock;
    struct block *block;
    
    block = block_for_pc(pc);
    if (block) {
	theBlock = NXMapGet(blocks, block);
	if (!theBlock)
	    theBlock = [self makeBlockForBlock: block];
    } else {
	static Block *emptyBlock = nil;
	
	if (!emptyBlock)
	    emptyBlock = [Block new];
	theBlock = emptyBlock;
    }
    
    return theBlock;
}

-(int)hasSymbolsForPC: (unsigned)pc
{
    return block_for_pc(pc) ? YES : NO;
}

static int getFramesUpTo(GdbController *self, int maxFrames, int *selFrame)
{
    int count, i;
    FRAME fr;
    Frame *frame;
    
    if (target_has_stack) {
	for (fr = get_current_frame(), self->numFrames = 0;
	     fr && (self->numFrames < self->maxFrames);
	     fr = get_prev_frame(fr), self->numFrames++);
	*self->selFrame = self->numFrames - 1;
	if (self->numFramesAlloced < self->numFrames) {
	    int numFramesToAlloc;
	    for (numFramesToAlloc = self->numFramesAlloced * 2;
	         numFramesToAlloc < self->numFrames;
		 numFramesToAlloc *= 2);
	    [self allocFrames: numFramesToAlloc];
	}
	for (fr = get_current_frame(),
	     i = self->numFrames,
	     frame = self->frames + i - 1;
             i--;
	     fr = get_prev_frame(fr), frame--) {
	    frame->frame = fr->frame;
	    frame->pc = fr->pc;
	    frame->block = nil;
	    if (fr == selected_frame)
		*self->selFrame = i; 
	}
	return self->numFrames;
    } else
	return 0;
}

-(int)getFramesUpTo: (int)mFrames selFrame: (int *)sFrame
{
    int wasError;
    maxFrames = mFrames;
    selFrame = sFrame;
    catch_errors_was_error(getFramesUpTo, self, &wasError);
    if (wasError) {
	printf("Error occured processing command from Edit.\n");
	[self programStopped];
	return 0;
    } else
	return numFrames;
}

-(List *)blockValuesFor: (void *)data
{
    struct block *block = data;
    int nSyms = block ? BLOCK_NSYMS(block) : 0, i;
    List *vals = [List newCount: nSyms];
    struct symbol *sym, *newSym;
    
    for (i = 0; i < nSyms; i++) {
	sym = BLOCK_SYM (block, i);
	newSym = lookup_symbol(SYMBOL_NAME(sym), block, VAR_NAMESPACE,
	                    NULL, NULL);
	if (sym == newSym)
	    [vals addObject: [self makeValue: sym]];
    }
    return vals;
}

-(Type *)getObjectTypeFor: (unsigned)objectAddress
{
    struct type *type;
    if (type = (struct type *)typeForObject(objectAddress))
	return [self makeType: type];
    else
	return nil;
}

-(char *)demangleName: (char *)name
{
    char *cplusName;
    if (cplusName = NXMapGet(demangledNames, name))
	return cplusName;
    else if (cplusName = cplus_demangle(name, DMGL_PARAMS | DMGL_ANSI)) {
	NXMapInsert(demangledNames, name, cplusName);
	return cplusName;
    } else
	return name;
}

-(char *)nameForPC: (unsigned)pc
{
    struct minimal_symbol *minSym = lookup_minimal_symbol_by_pc(pc);
    return minSym ? [self demangleName: minSym->name] : "??";
}

-(void)setClient: (id<GdbUser>)theClient
{
    char *execName, *lastSlash;
    NXProxy *interruptHandler;
    
    mutex_lock(clientLock);
    client = theClient;
    if (execName = (char *)get_exec_file(0)) {
	lastSlash = (char *)rindex(execName, '/');
	if (lastSlash)
	    execName = lastSlash + 1;
    }
NX_DURING
    [client setGdbName: execName];
    interruptHandler = [interruptConnection getLocal: self];
    [client setInterruptHandler: interruptHandler];
NX_HANDLER
    badRemoteMethod(&NXLocalHandler, sel_getName(_cmd));
NX_ENDHANDLER
    condition_signal(clientCond);
    mutex_unlock(clientLock);
}

-(void)openFile: (char *)fileName fromTrueLine: (int)firstLine to: (int)lastLine
{
NX_DURING
   [client openFile: fileName onHost: gdbHost
	fromTrueLine: firstLine to: lastLine];
NX_HANDLER
    badRemoteMethod(&NXLocalHandler, sel_getName(_cmd));
NX_ENDHANDLER
}

-(void)openFile: (char *)fileName atTrueLine: (int)line
{
NX_DURING
   [client openFile: fileName onHost: gdbHost atTrueLine: line];
NX_HANDLER
   badRemoteMethod(&NXLocalHandler, sel_getName(_cmd));
NX_ENDHANDLER
}

-(void)errorOpeningFile: (char *)filename fromLine: (int)start to: (int)end
{
    if (start == end)
	printf("Error opening %s at %d.\n", filename, start);
    else
	printf("Error opening %s from %d to %d.\n", filename, start, end);
}

-(void)programStopped
{
    lastToldAboutRunning = NO;
NX_DURING
    [client programStopped];
NX_HANDLER
    badRemoteMethod(&NXLocalHandler, sel_getName(_cmd));
NX_ENDHANDLER
}

-(void)programIsDead
{
    lastToldAboutRunning = NO;
NX_DURING
    [client programIsDead];
NX_HANDLER
    badRemoteMethod(&NXLocalHandler, sel_getName(_cmd));
NX_ENDHANDLER
}

-(void)frameChanged;
{
NX_DURING
    [client frameChanged];
NX_HANDLER
    badRemoteMethod(&NXLocalHandler, sel_getName(_cmd));
NX_ENDHANDLER
}

-(void)programIsRunning
{
    if (!lastToldAboutRunning) {
	lastToldAboutRunning = YES;
NX_DURING
	[client programIsRunning];
NX_HANDLER
	badRemoteMethod(&NXLocalHandler, sel_getName(_cmd));
NX_ENDHANDLER
    }
}

void
tell_client_program_stopped()
{
    if (controller) {
NX_DURING
	[(cthread_self() == gdb_thread) ? remote_controller : controller programStopped];
NX_HANDLER
	badRemoteMethod(&NXLocalHandler, "tell_client_program_stopped");
NX_ENDHANDLER
    }
}

void
tell_client_program_is_running()
{
    if (controller) {
NX_DURING
	[(cthread_self() == gdb_thread) ? remote_controller : controller programIsRunning];
NX_HANDLER
	badRemoteMethod(&NXLocalHandler, "tell_client_program_stopped");
NX_ENDHANDLER
    }
}

void
tell_client_program_is_dead()
{
    if (controller) {
NX_DURING
	[(cthread_self() == gdb_thread) ? remote_controller : controller programIsDead];
NX_HANDLER
	badRemoteMethod(&NXLocalHandler, "tell_client_program_is_dead");
NX_ENDHANDLER
    }
}

void
tell_client_frame_changed()
{
    if (controller) {
NX_DURING
	[(cthread_self() == gdb_thread) ? remote_controller : controller frameChanged];
NX_HANDLER
	badRemoteMethod(&NXLocalHandler, "tell_client_program_stopped");
NX_ENDHANDLER
    }
}

@end