#import "Value.h"
#import <objc/HashTable.h>

static id <DataGetter>dataGetter;
static id <ErrorDelegate>errorDelegate;
static HashTable *valueCasts, *objectCasts;

char *valueErrorStrings[NUM_ERROR_CODES] = {
    "No error",
    "A source for the data for the value is needed",
    "The value must have a relative location",
    "The value must be a scaler",
    "The value must be a float",
    "The value must be an enum",
    "The value must be an aggregate",
    "The value must be a pointer",
    "The value must be a string"
    "The value must be a function"
};

@implementation Value
+(void)setDataGetter: (id <DataGetter>)theGetter { dataGetter = theGetter; }
+(id <DataGetter>)dataGetter		{ return dataGetter; }
+(void)setErrorDelegate: (id <ErrorDelegate>)theDelegate
{ errorDelegate = theDelegate; }
+(id <ErrorDelegate>)errorDelegate	{ return errorDelegate; }
-(void)setValueName: (char *)theName	{ valueName = theName; }
-(char *)valueName			{ return valueName; }
-(void)setType: (Type *)theType		{ _type = theType; }
-(void)setLocType: (LocType)locType	{ location.locType = locType; }
-(LocType)locType			{ return location.locType; }
-(void)setLocPlace: (int)locPlace	{ location.locPlace.loc = locPlace; }
-(int)locPlace				{ return location.locPlace.loc; }

-(Type *)type
{
    Type *castType;
    
    castType = [objectCasts valueForKey: self];
    if (!castType)
	castType = [valueCasts valueForKey: self];
    return castType ? castType : _type;
}

+initialize
{
    valueCasts = [HashTable new];
    objectCasts = [HashTable new];
    return self;
}

BOOL static inline needsMore(Location location)
{
    return (location.locType <= LTYPE_MAX_INDEPENDENT);
}

-(unsigned)size
{
    return [[self type] size];
}

-(unsigned)scalerValueFrom: (unsigned)from
{
    unsigned uVal;
    Type *type = [self type];

    if ([type isScaler]) {
	uVal = [dataGetter getScalerValueFrom: from
	                                  at: &location
			            withSize: [type size]];
	if ([type isObject]) {
	    Type *objectType = [dataGetter getObjectTypeFor: uVal];
	    if (objectType) {
		type = objectType;
		[objectCasts insertKey: self value: objectType];
	    } else
		[objectCasts removeKey: self];
	}
	if ((location.locType == LTYPE_RELATIVE_BIT)
	    && (location.locPlace.locBits.bitSize)) {
	    int lShift = location.locPlace.locBits.bitPos % 32,
	        rShift = 32 - location.locPlace.locBits.bitSize;
	    
	    if ([(Integer *)type isSigned]) {
		((int)uVal) <<= lShift;
		((int)uVal) >>= rShift;
	    } else {
		uVal <<= lShift;
		uVal >>= rShift;
	    }
	}
	return uVal;
    } else {
	[errorDelegate valueError: ERROR_NOT_SCALER];
	return 0;
    }
}

-(float)floatValueFrom: (unsigned)from
{
    Type *type = [self type];
    if ([type isFloat] || [type isDouble]) {
	return (float)[dataGetter getDoubleValueFrom: from at: &location withSize: [type size]];
    } else {
	[errorDelegate valueError: ERROR_NOT_FLOAT];
	return 0.0;
    }
}

-(char *)enumValueFrom: (unsigned)from
{
    Type *type = [self type];
    unsigned lValue;
    if ([type isEnum]) {
	lValue = [dataGetter getScalerValueFrom: from
	                                     at: &location
				       withSize: [type size]];
	return [[(Enumeration *)type valFor: lValue] name];
    } else {
	[errorDelegate valueError: ERROR_NOT_ENUM];
	return NULL;
    }
}

-(char *)stringValueFrom: (unsigned)from
{
    Type *type = [self type];
    if ([type isString]) {
	if ([type isPointer])
	    return [dataGetter getStringValueFrom: from at: &location];
	else
	    return [dataGetter getStringValueForArrayFrom: from at: &location];
    } else {
	[errorDelegate valueError: ERROR_NOT_STRING];
	return NULL;
    }
}

-(char *)functionValueFrom: (unsigned)from
{
    if ([[self type] isFunc]) {
	return [dataGetter getFunctionNameFrom: from at: &location];
    } else {
	[errorDelegate valueError: ERROR_NOT_FUNC];
	return NULL;
    }
}

-(int)numSubValues
{
    Type *type = [self type];
    unsigned flags = [type flags];

    if (flags & (ISAGGREGATE | ISPOINTER))
	return [(AggregateType *)type numSubValues];
    else {
	[errorDelegate valueError: ERROR_NOT_AGGREGATE];
	return 0;
    }
}

-(Value *)subValueAt: (int)nValue
{
    Type *type = [self type];
    unsigned flags = [type flags];
    
    if (flags & (ISAGGREGATE | ISPOINTER))
	return [(AggregateType *)type subValueAt: nValue];
    else {
	[errorDelegate valueError: ERROR_NOT_AGGREGATE];
	return nil;
    }
}

-(Value *)subValueAt: (int)nValue
           withFlags: (out unsigned *)pFlags
	     andName: (out char **)pName
{
    Value *value = [self subValueAt: nValue];
    *pFlags = [value flags];
    *pName  = [value valueName];
    return value;
}

-(Value *)subValueNamed: (char *)name index: (int *)pIndex
{
    int numSubValues = [self numSubValues], i;
    Value *subValue, *foundValue = nil;
    
   for (i = 0; !foundValue && (i < numSubValues); i++) {
	subValue = [self subValueAt: i];
	if (strcmp(name, [subValue valueName]) == 0)
	    foundValue = subValue;
    }
    if (foundValue)
	*pIndex = i - 1;
    return foundValue;
}

-(Value *)subValueFor: (Value *)oldValue
                 from: (int *)from
		index: (int *)index
{
    Value *value = [self subValueAt: *index];
    
    if (value != oldValue)
	value = [self subValueNamed: [oldValue valueName] index: index];
	
    if (value) {
	int isValid;
	
	*from = [value addressFrom: *from isValid: &isValid];
	if (!isValid)
	    value = nil;
    }
	
    return value;
}
	
-(unsigned)addressFrom: (unsigned)from
{
    if ([[self type] isPointer])
	return [self scalerValueFrom: from];
    else
	return [dataGetter getAddressFrom: from at: &location];
}

-(unsigned)addressFrom: (unsigned)from isValid: (int *)isValid
{
    Type *type = [self type];
    unsigned address;
    
    if ([type isPointer]) {
	address = [self scalerValueFrom: from];
	*isValid = [dataGetter isDataFrom: address
	                              for: [(Pointer *)type indSize]];
    } else {
	address = [dataGetter getAddressFrom: from at: &location];
	*isValid = [dataGetter isDataFrom: address for: [type size]];
    }
    return address;
}

-(unsigned)flags
{
    return [[self type] flags];
}
    
-(char *)typeName
{
    return [[self type] typeName];
}

-(int)castToTypeNamed: (char *)typeName from: (unsigned)from
{
    Type *newType;
    
    if (newType = [Type getTypeNamed: typeName from: from]) {
	[valueCasts insertKey: self value: newType];
	return YES;
    } else
	return NO;
}

@end